@@ -1,5 +1,6 @@
 #!/usr/bin/env python3
-# -*- coding: utf-8 -*-
+
+from __future__ import annotations
 
 import re
 import logging
@@ -9,15 +10,17 @@ from email import policy, message_from_bytes
 from email.message import EmailMessage
 from io import BytesIO
 from pathlib import Path
-from typing import Union, List, Tuple, Dict, cast, Any, Optional
+from typing import Any
 
 from extract_msg import openMsg
-from extract_msg.message import Message as MsgObj
+from extract_msg.msg_classes import MessageBase
+from extract_msg.attachments import AttachmentBase, SignedAttachment
+from extract_msg.properties import FixedLengthProp
 from RTFDE.exceptions import MalformedEncapsulatedRtf, NotEncapsulatedRtf  # type: ignore
 from RTFDE.deencapsulate import DeEncapsulator  # type: ignore
 from oletools.common.codepages import codepage2codec  # type: ignore
 
-from ..exceptions import InvalidMISPObject, PyMISPNotImplementedYet, MISPObjectException, NewAttributeError
+from ..exceptions import InvalidMISPObject, MISPObjectException, NewAttributeError
 from .abstractgenerator import AbstractMISPObjectGenerator
 
 logger = logging.getLogger('pymisp')
@@ -28,15 +31,14 @@ class MISPMsgConverstionError(MISPObjectException):
 
 
 class EMailObject(AbstractMISPObjectGenerator):
-    def __init__(self, filepath: Optional[Union[Path, str]]=None, pseudofile: Optional[BytesIO]=None,
-                 attach_original_email: bool = True, **kwargs):
+    def __init__(self, filepath: Path | str | None=None, pseudofile: BytesIO | bytes | None=None,  # type: ignore[no-untyped-def]
+                 attach_original_email: bool = True, **kwargs) -> None:
         super().__init__('email', **kwargs)
 
         self.attach_original_email = attach_original_email
-        self.encapsulated_body: Union[str, None] = None
-        self.eml_from_msg: Union[bool, None] = None
-        self.raw_emails: Dict[str, Union[BytesIO, None]] = {'msg': None,
-                                                            'eml': None}
+        self.encapsulated_body: str | None = None
+        self.eml_from_msg: bool | None = None
+        self.raw_emails: dict[str, BytesIO | None] = {'msg': None, 'eml': None}
 
         self.__pseudofile = self.create_pseudofile(filepath, pseudofile)
         self.email = self.parse_email()
@@ -48,7 +50,6 @@ class EMailObject(AbstractMISPObjectGenerator):
         eml = message_from_bytes(content_in_bytes,
                                  _class=EmailMessage,
                                  policy=policy.default)
-        eml = cast(EmailMessage, eml)  # Only needed to quiet mypy
         if len(eml) != 0:
             self.raw_emails['eml'] = self.__pseudofile
             return eml
@@ -65,23 +66,22 @@ class EMailObject(AbstractMISPObjectGenerator):
                     return message
         except ValueError as _e:  # Exception
             logger.debug("Email not in .msg format or is a corrupted .msg. Attempting to decode email from other formats.")
-            logger.debug("Error: {} ".format(_e))
+            logger.debug(f"Error: {_e} ")
         try:
             if content_in_bytes[:3] == b'\xef\xbb\xbf':  # utf-8-sig byte-order mark (BOM)
                 eml_bytes = content_in_bytes.decode("utf_8_sig").encode("utf-8")
                 eml = email.message_from_bytes(eml_bytes,
                                                policy=policy.default)
-                eml = cast(EmailMessage, eml)  # Only needed to quiet mypy
                 if len(eml) != 0:
                     self.raw_emails['eml'] = BytesIO(eml_bytes)
                     return eml
         except UnicodeDecodeError:
             pass
-        raise PyMISPNotImplementedYet("EmailObject does not know how to decode data passed to it. Object may not be an email. If this is an email please submit it as an issue to PyMISP so we can add support.")
+        raise InvalidMISPObject("EmailObject does not know how to decode data passed to it. Object may not be an email. If this is an email please submit it as an issue to PyMISP so we can add support.")
 
     @staticmethod
-    def create_pseudofile(filepath: Optional[Union[Path, str]] = None,
-                          pseudofile: Optional[BytesIO] = None) -> BytesIO:
+    def create_pseudofile(filepath: Path | str | None = None,
+                          pseudofile: BytesIO | bytes | None = None) -> BytesIO:
         """Creates a pseudofile using directly passed data or data loaded from file path.
         """
         if filepath:
@@ -89,20 +89,23 @@ class EMailObject(AbstractMISPObjectGenerator):
                 return BytesIO(f.read())
         elif pseudofile and isinstance(pseudofile, BytesIO):
             return pseudofile
+        elif pseudofile and isinstance(pseudofile, bytes):
+            return BytesIO(pseudofile)
         else:
             raise InvalidMISPObject('File buffer (BytesIO) or a path is required.')
 
     def _msg_to_eml(self, msg_bytes: bytes) -> EmailMessage:
         """Converts a msg into an eml."""
-        msg_obj = openMsg(msg_bytes)
+        # NOTE: openMsg returns a MessageBase, not a MSGFile
+        msg_obj: MessageBase = openMsg(msg_bytes)
         # msg obj stores the original raw header here
         message, body, attachments = self._extract_msg_objects(msg_obj)
         eml = self._build_eml(message, body, attachments)
         return eml
 
-    def _extract_msg_objects(self, msg_obj: MsgObj) -> Tuple[EmailMessage, Dict, List[Any]]:
+    def _extract_msg_objects(self, msg_obj: MessageBase) -> tuple[EmailMessage, dict[str, Any], list[AttachmentBase] | list[SignedAttachment]]:
         """Extracts email objects needed to construct an eml from a msg."""
-        message: EmailMessage = email.message_from_string(msg_obj.header.as_string(), policy=policy.default)  # type: ignore
+        message: EmailMessage = email.message_from_string(msg_obj.header.as_string(), policy=policy.default)
         body = {}
         if msg_obj.body is not None:
             body['text'] = {"obj": msg_obj.body,
@@ -111,8 +114,11 @@ class EMailObject(AbstractMISPObjectGenerator):
                             "cte": "base64"}
         if msg_obj.htmlBody is not None:
             try:
-                _html_encoding_raw = msg_obj.props['3FDE0003'].value
-                _html_encoding = codepage2codec(_html_encoding_raw)
+                if isinstance(msg_obj.props['3FDE0003'], FixedLengthProp):
+                    _html_encoding_raw = msg_obj.props['3FDE0003'].value
+                    _html_encoding = codepage2codec(_html_encoding_raw)
+                else:
+                    _html_encoding = msg_obj.stringEncoding
             except KeyError:
                 _html_encoding = msg_obj.stringEncoding
             body['html'] = {'obj': msg_obj.htmlBody.decode(),
@@ -145,17 +151,16 @@ class EMailObject(AbstractMISPObjectGenerator):
         attachments = msg_obj.attachments
         return message, body, attachments
 
-    def _build_eml(self, message: EmailMessage, body: dict, attachments: list) -> EmailMessage:
+    def _build_eml(self, message: EmailMessage, body: dict[str, Any], attachments: list[Any]) -> EmailMessage:
         """Constructs an eml file from objects extracted from a msg."""
         # Order the body objects by increasing complexity and toss any missing objects
-        body_objects: List[dict] = [body.get('text', {}),
-                                    body.get('html', {}),
-                                    body.get('rtf', {})]
-        body_objects = [i for i in body_objects if i != {}]
+        body_objects: list[dict[str, Any]] = [i for i in [body.get('text'),
+                                                          body.get('html'),
+                                                          body.get('rtf')] if i is not None]
         # If this a non-multipart email then we only need to attach the payload
         if message.get_content_maintype() != 'multipart':
             for _body in body_objects:
-                if "text/{0}".format(_body['subtype']) == message.get_content_type():
+                if "text/{}".format(_body['subtype']) == message.get_content_type():
                     message.set_content(**_body)
                     return message
             raise MISPMsgConverstionError("Unable to find appropriate eml payload in message body.")
@@ -167,8 +172,8 @@ class EMailObject(AbstractMISPObjectGenerator):
         if isinstance(body.get('html', None), dict):
             _html = body.get('html', {}).get('obj')
             for attch in attachments:
-                if _html.find("cid:{0}".format(attch.cid)) != -1:
-                    _content_type = attch._getStringStream('__substg1.0_370E')
+                if _html.find(f"cid:{attch.cid}") != -1:
+                    _content_type = attch.getStringStream('__substg1.0_370E')
                     maintype, subtype = _content_type.split("/", 1)
                     related_content[attch.cid] = (attch,
                                                   {'obj': attch.data,
@@ -189,11 +194,19 @@ class EMailObject(AbstractMISPObjectGenerator):
             for mime_items in related_content.values():
                 if isinstance(mime_items[1], dict):
                     message.add_related(**mime_items[1])
-                cur_attach = message.get_payload()[-1]
+                if p := message.get_payload():
+                    if isinstance(p, list):
+                        cur_attach = p[-1]
+                    else:
+                        cur_attach = p
                 self._update_content_disp_properties(mime_items[0], cur_attach)
             if body.get('text', None):
                 # Now add the HTML as an alternative within the related obj
-                related = message.get_payload()[0]
+                if p := message.get_payload():
+                    if isinstance(p, list):
+                        related = p[0]
+                    else:
+                        related = p
                 related.add_alternative(**body.get('html'))
         else:
             for mime_dict in body_objects:
@@ -205,20 +218,25 @@ class EMailObject(AbstractMISPObjectGenerator):
                     message.add_alternative(**mime_dict)
         for attch in attachments:  # Add attachments at the end.
             if attch.cid not in related_content.keys():
-                _content_type = attch._getStringStream('__substg1.0_370E')
+                _content_type = attch.getStringStream('__substg1.0_370E')
                 maintype, subtype = _content_type.split("/", 1)
                 message.add_attachment(attch.data,
                                        maintype=maintype,
                                        subtype=subtype,
                                        cid=attch.cid,
                                        filename=attch.longFilename)
-                cur_attach = message.get_payload()[-1]
+                if p := message.get_payload():
+                    if isinstance(p, list):
+                        cur_attach = p[-1]
+                    else:
+                        cur_attach = p
                 self._update_content_disp_properties(attch, cur_attach)
-        message.set_boundary(_orig_boundry)  # Set back original boundary
+        if _orig_boundry is not None:
+            message.set_boundary(_orig_boundry)  # Set back original boundary
         return message
 
     @staticmethod
-    def _update_content_disp_properties(msg_attch, eml_attch):
+    def _update_content_disp_properties(msg_attch: AttachmentBase, eml_attch: EmailMessage) -> None:
         """Set Content-Disposition params on binary eml objects
 
         You currently have to set non-filename content-disp params by hand in python.
@@ -228,18 +246,18 @@ class EMailObject(AbstractMISPObjectGenerator):
         for num, name in attch_cont_disp_props.items():
             try:
                 eml_attch.set_param(name,
-                                    email.utils.format_datetime(msg_attch.props[num].value),
+                                    email.utils.format_datetime(msg_attch.props.getValue(num)),
                                     header='Content-Disposition')
             except KeyError:
                 # It's fine if they don't have those values
                 pass
 
     @property
-    def attachments(self) -> List[Tuple[str, BytesIO]]:
+    def attachments(self) -> list[tuple[str | None, BytesIO]]:
         to_return = []
         try:
             for attachment in self.email.iter_attachments():
-                content = attachment.get_content()  # type: ignore
+                content = attachment.get_content()
                 if isinstance(content, str):
                     content = content.encode()
                 to_return.append((attachment.get_filename(), BytesIO(content)))
@@ -249,7 +267,7 @@ class EMailObject(AbstractMISPObjectGenerator):
             pass
         return to_return
 
-    def generate_attributes(self):
+    def generate_attributes(self) -> None:
 
         # Attach original & Converted
         if self.attach_original_email is not None:
@@ -261,21 +279,30 @@ class EMailObject(AbstractMISPObjectGenerator):
                                    data=self.raw_emails.get('msg'))
 
         message = self.email
+        body: EmailMessage
 
-        for _pref, body in message._find_body(message, preferencelist=['plain', 'html']):
-            comment = "{0} body".format(body.get_content_type())
+        if body := message.get_body(preferencelist=['plain']):
+            comment = f"{body.get_content_type()} body"
             if self.encapsulated_body == body.get_content_type():
                 comment += " De-Encapsulated from RTF in original msg."
             self.add_attribute("email-body",
                                body.get_content(),
                                comment=comment)
 
-        headers = ["{}: {}".format(k, v) for k, v in message.items()]
+        if body := message.get_body(preferencelist=['html']):
+            comment = f"{body.get_content_type()} body"
+            if self.encapsulated_body == body.get_content_type():
+                comment += " De-Encapsulated from RTF in original msg."
+            self.add_attribute("email-body",
+                               body.get_content(),
+                               comment=comment)
+
+        headers = [f"{k}: {v}" for k, v in message.items()]
         if headers:
             self.add_attribute("header", "\n".join(headers))
 
-        if "Date" in message and message.get('date').datetime is not None:
-            self.add_attribute("send-date", message.get('date').datetime)
+        if "Date" in message and message['date'].datetime is not None:
+            self.add_attribute("send-date", message['date'].datetime)
 
         if "To" in message:
             self.__add_emails("to", message["To"])
@@ -319,61 +346,120 @@ class EMailObject(AbstractMISPObjectGenerator):
 
         self.__generate_received()
 
-    def __add_emails(self, typ: str, data: str, insert_display_names: bool = True):
-        addresses = []
-        display_names = []
+    def __add_emails(self, typ: str, data: str, insert_display_names: bool = True) -> None:
+        addresses: list[dict[str, str]] = []
+        display_names: list[dict[str, str]] = []
 
         for realname, address in email.utils.getaddresses([data]):
             if address and realname:
-                addresses.append({"value": address, "comment": "{} <{}>".format(realname, address)})
+                addresses.append({"value": address, "comment": f"{realname} <{address}>"})
             elif address:
                 addresses.append({"value": address})
             else:  # parsing failed, skip
                 continue
 
             if realname:
-                display_names.append({"value": realname, "comment": "{} <{}>".format(realname, address)})
+                display_names.append({"value": realname, "comment": f"{realname} <{address}>"})
 
-        if addresses:
-            self.add_attributes(typ, *addresses)
+        for a in addresses:
+            self.add_attribute(typ, **a)
         if insert_display_names and display_names:
             try:
-                self.add_attributes("{}-display-name".format(typ), *display_names)
+                for d in display_names:
+                    self.add_attribute(f"{typ}-display-name", **d)
             except NewAttributeError:
                 # email object doesn't support display name for all email addrs
                 pass
 
-    def __generate_received(self):
-        """
-        Extract IP addresses from received headers that are not private. Also extract hostnames or domains.
-        """
-        received_items = self.email.get_all("received")
-        if received_items is None:
-            return
-        for received in received_items:
-            fromstr = re.split(r"\sby\s", received)[0].strip()
-            if fromstr.startswith('from') is not True:
-                continue
-            for i in ['(', ')', '[', ']']:
-                fromstr = fromstr.replace(i, " ")
-            tokens = fromstr.split(" ")
-            ip = None
-            for token in tokens:
-                try:
-                    ip = ipaddress.ip_address(token)
-                    break
-                except ValueError:
-                    pass  # token is not IP address
+    def extract_matches(self, pattern: re.Pattern[str], text: str) -> list[tuple[str, ...]]:
+        """Returns all regex matches for a given pattern in a text."""
+        return re.findall(pattern, text)
 
-            if not ip or ip.is_private:
-                continue  # skip header if IP not found or is private
-
-            self.add_attribute("received-header-ip", value=str(ip), comment=fromstr)
-
-        # The hostnames and/or domains always come after the "Received: from"
-        # part so we can use regex to pick up those attributes.
-        received_from = re.findall(r'(?<=from\s)[\w\d\.\-]+\.\w{2,24}', str(received_items))
+    def add_ip_attribute(self, ip_candidate: str, received: str, seen_attributes: set[tuple[str, str]]) -> None:
+        """Validates and adds an IP address to MISP if it's public and not already seen during extraction."""
         try:
-            [self.add_attribute("received-header-hostname", i) for i in received_from]
-        except Exception:
-            pass
+            ip = ipaddress.ip_address(ip_candidate)
+            if not ip.is_private and ("received-header-ip", ip_candidate) not in seen_attributes:
+                self.add_attribute("received-header-ip", ip_candidate, comment=received)
+                seen_attributes.add(("received-header-ip", ip_candidate))
+        except ValueError:
+            pass  # Invalid IPs are ignored
+
+    def add_hostname_attribute(self, hostname: str, received: str, seen_attributes: set[tuple[str, str]]) -> None:
+        """Validates and adds a hostname to MISP if it contains a valid TLD-like format and is not already seen."""
+        if "." in hostname and not hostname.endswith(".") and len(hostname.split(".")[-1]) > 1:
+            if ("received-header-hostname", hostname) not in seen_attributes:
+                self.add_attribute("received-header-hostname", hostname, comment=received)
+                seen_attributes.add(("received-header-hostname", hostname))
+
+    def process_received_header(self, received: str, seen_attributes: set[tuple[str, str]]) -> None:
+        """Processes a single 'Received' header and extracts hostnames and IPs."""
+
+        # Regex patterns
+        received_from_regex = re.compile(
+            r'from\s+([\w.-]+)'  # Declared sending hostname
+            r'(?:\s+\(([^)]+)\))?'  # Reverse DNS hostname inside parentheses
+        )
+        ipv4_regex = re.compile(
+            r'\[(?P<ipv4_brackets>(?:25[0-5]|2[0-4][0-9]|1[0-9]{2}|[1-9]?[0-9])\.'
+            r'(?:25[0-5]|2[0-4][0-9]|1[0-9]{2}|[1-9]?[0-9])\.'
+            r'(?:25[0-5]|2[0-4][0-9]|1[0-9]{2}|[1-9]?[0-9])\.'
+            r'(?:25[0-5]|2[0-4][0-9]|1[0-9]{2}|[1-9]?[0-9]))\]'  # IPv4 inside []
+            r'|\((?P<ipv4_parentheses>(?:25[0-5]|2[0-4][0-9]|1[0-9]{2}|[1-9]?[0-9])\.'
+            r'(?:25[0-5]|2[0-4][0-9]|1[0-9]{2}|[1-9]?[0-9])\.'
+            r'(?:25[0-5]|2[0-4][0-9]|1[0-9]{2}|[1-9]?[0-9])\.'
+            r'(?:25[0-5]|2[0-4][0-9]|1[0-9]{2}|[1-9]?[0-9]))\)'  # IPv4 inside ()
+            r'|(?<=\.\s)(?P<ipv4_after_domain>(?:25[0-5]|2[0-4][0-9]|1[0-9]{2}|[1-9]?[0-9])\.'
+            r'(?:25[0-5]|2[0-4][0-9]|1[0-9]{2}|[1-9]?[0-9])\.'
+            r'(?:25[0-5]|2[0-4][0-9]|1[0-9]{2}|[1-9]?[0-9])\.'
+            r'(?:25[0-5]|2[0-4][0-9]|1[0-9]{2}|[1-9]?[0-9]))\b'  # IPv4 appearing after a domain.
+        )
+        ipv6_regex = re.compile(
+            r'\b(?:[a-fA-F0-9]{1,4}:[a-fA-F0-9]{1,4}(?::[a-fA-F0-9]{1,4}){0,6})\b'
+        )
+
+        # Extract hostnames
+        matches = self.extract_matches(received_from_regex, received)
+        for match in matches:
+            declared_sending_host = match[0].strip() if match[0] else None
+            reverse_dns_host = match[1].split()[0].strip("[]()").rstrip('.') if match[1] else None
+
+            if declared_sending_host:
+                clean_host = declared_sending_host.strip("[]()")
+                try:
+                    ipaddress.ip_address(declared_sending_host)
+                    self.add_ip_attribute(declared_sending_host, received, seen_attributes)
+                except ValueError:
+                    self.add_hostname_attribute(declared_sending_host, received, seen_attributes)
+
+            if reverse_dns_host:
+                try:
+                    ipaddress.ip_address(reverse_dns_host)
+                    self.add_ip_attribute(reverse_dns_host, received, seen_attributes)
+                except ValueError:
+                    self.add_hostname_attribute(reverse_dns_host, received, seen_attributes)
+
+        # Extract and add **only valid** IPv4 addresses
+        for ipv4_match in self.extract_matches(ipv4_regex, received):
+            ip_candidate = ipv4_match[0] or ipv4_match[1] or ipv4_match[2]  # Select first non-empty match
+            if ip_candidate:
+                self.add_ip_attribute(ip_candidate, received, seen_attributes)
+
+        # Extract and add IPv6 addresses
+        for ipv6_match in self.extract_matches(ipv6_regex, received):
+            self.add_ip_attribute(ipv6_match, received, seen_attributes)
+
+    def __generate_received(self) -> None:
+        """
+        Extracts public IP addresses and hostnames from "Received" email headers.
+        """
+
+        received_items = self.email.get_all("Received")
+        if not received_items:
+            return
+
+        # Track added attributes to prevent duplicates (store as (type, value) tuples)
+        seen_attributes: set[tuple[str, str]] = set()
+
+        for received in received_items:
+            self.process_received_header(received, seen_attributes)
